#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["get_pipelines_config", "transform_dependencies"]

from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from rialto.common.utils import load_yaml


class IntervalConfig(BaseModel):
    units: str
    value: int


class ScheduleConfig(BaseModel):
    frequency: str
    day: Optional[int] = 0
    info_date_shift: Union[Optional[IntervalConfig], List[IntervalConfig]] = IntervalConfig(units="days", value=0)


class DependencyConfig(BaseModel):
    table: str
    name: Optional[str] = None
    date_col: Optional[str] = None
    interval: IntervalConfig


class ModuleConfig(BaseModel):
    python_module: str
    python_class: str


class MailConfig(BaseModel):
    sender: str
    to: List[str]
    smtp: str
    subject: str
    sent_empty: Optional[bool] = False


class GeneralConfig(BaseModel):
    target_schema: str
    target_partition_column: str
    source_date_column_property: Optional[str] = None
    watched_period_units: str
    watched_period_value: int
    job: str
    mail: MailConfig


class PipelineConfig(BaseModel):
    name: str
    module: Optional[ModuleConfig] = None
    schedule: ScheduleConfig
    dependencies: List[DependencyConfig] = []


class PipelinesConfig(BaseModel):
    general: GeneralConfig
    pipelines: list[PipelineConfig]


def get_pipelines_config(path) -> PipelinesConfig:
    """Load and parse yaml config"""
    return PipelinesConfig(**load_yaml(path))


def transform_dependencies(dependencies: List[DependencyConfig]) -> Dict:
    """Transform dependency config list into a dictionary"""
    res = {}
    for dep in dependencies:
        if dep.name:
            res[dep.name] = dep
    return res



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["disable_job_decorators"]

import importlib
import typing
from contextlib import contextmanager
from unittest.mock import patch


def _passthrough_decorator(x: typing.Callable) -> typing.Callable:
    if type(x) is str:
        return _passthrough_decorator

    else:
        return x


@contextmanager
def _disable_job_decorators() -> None:
    patches = [
        patch("rialto.jobs.decorators.datasource", _passthrough_decorator),
        patch("rialto.jobs.decorators.decorators.datasource", _passthrough_decorator),
        patch("rialto.jobs.decorators.job", _passthrough_decorator),
        patch("rialto.jobs.decorators.decorators.job", _passthrough_decorator),
    ]

    for i in patches:
        i.start()

    yield

    for i in patches:
        i.stop()


@contextmanager
def disable_job_decorators(module) -> None:
    """
    Disables job decorators in a python module. Useful for testing your rialto jobs and datasources.

    :param module: python module with the decorated functions.
    :return: None
    """
    with _disable_job_decorators():
        importlib.reload(module)
        yield

    importlib.reload(module)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["datasource", "job"]

import importlib.metadata
import inspect
import typing

from loguru import logger

from rialto.jobs.decorators.job_base import JobBase
from rialto.jobs.decorators.resolver import Resolver


def datasource(ds_getter: typing.Callable) -> typing.Callable:
    """
    Dataset reader functions decorator.

    Registers a data-reading function into a rialto job prerequisite.
    You can then request the job via job function arguments.

    :param ds_getter:  dataset reader function
    :return: raw reader function, unchanged
    """
    Resolver.register_callable(ds_getter)
    return ds_getter


def _get_module(stack: typing.List) -> typing.Any:
    last_stack = stack[1]
    mod = inspect.getmodule(last_stack[0])
    return mod


def _get_version(module: typing.Any) -> str:
    try:
        parent_name, _, _ = module.__name__.partition(".")
        return importlib.metadata.version(parent_name)

    except Exception:
        logger.warning(f"Failed to get library {module.__name__} version!")
        return "N/A"


def _generate_rialto_job(callable: typing.Callable, module: object, class_name: str, version: str) -> typing.Type:
    generated_class = type(
        class_name,
        (JobBase,),
        {
            "get_custom_callable": lambda _: callable,
            "get_job_version": lambda _: version,
            "get_job_name": lambda _: class_name,
        },
    )

    generated_class.__module__ = module.__name__
    setattr(module, class_name, generated_class)

    logger.info(f"Registered {class_name} in {module}")
    return generated_class


def job(name_or_callable: typing.Union[str, typing.Callable]) -> typing.Union[typing.Callable, typing.Type]:
    """
    Rialto jobs decorator.

    Transforms a python function into a rialto transormation, which can be imported and ran by Rialto Runner.
    Allows a custom name, via @job("custom_name_here") or can be just used as @job and the function's name is used.

    :param name_or_callable:  str for custom job name. Otherwise, run function.
    :return: One more job wrapper for run function (if custom name specified).
             Otherwise, generates Rialto Transformation Type and returns it for in-module registration.
    """
    stack = inspect.stack()

    module = _get_module(stack)
    version = _get_version(module)

    if type(name_or_callable) is str:

        def inner_wrapper(callable):
            return _generate_rialto_job(callable, module, name_or_callable, version)

        return inner_wrapper

    else:
        name = name_or_callable.__name__
        return _generate_rialto_job(name_or_callable, module, name, version)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["DatabricksLoader"]

from datetime import date

from pyspark.sql import DataFrame, SparkSession

from rialto.common.table_reader import TableReader
from rialto.loader.interfaces import DataLoader


class DatabricksLoader(DataLoader):
    """Implementation of DataLoader using TableReader to access feature tables"""

    def __init__(self, spark: SparkSession, schema: str, date_column: str = "INFORMATION_DATE"):
        super().__init__()

        self.reader = TableReader(spark)
        self.schema = schema
        self.date_col = date_column

    def read_group(self, group: str, information_date: date) -> DataFrame:
        """
        Read a feature group by getting the latest partition by date

        :param group: group name
        :param information_date: partition date
        :return: dataframe
        """
        return self.reader.get_latest(
            f"{self.schema}.{group}", until=information_date, date_column=self.date_col, uppercase_columns=True
        )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["ConfigException", "FeatureStoreConfig", "ConfigHolder"]

import datetime
import typing

from pydantic import BaseModel


class ConfigException(Exception):
    """Wrong Configuration Exception"""

    pass


class FeatureStoreConfig(BaseModel):
    """Configuration of Feature Store Paths"""

    feature_store_schema: str = None
    feature_metadata_schema: str = None


class ConfigHolder:
    """
    Main Rialto Jobs config holder.

    Configured via job_runner and then called from job_base / job decorators.
    """

    _config = {}
    _dependencies = {}
    _run_date = None
    _feature_store_config: FeatureStoreConfig = None

    @classmethod
    def set_run_date(cls, run_date: datetime.date) -> None:
        """
        Inicialize run Date

        :param run_date: datetime.date, run date
        :return: None
        """
        cls._run_date = run_date

    @classmethod
    def get_run_date(cls) -> datetime.date:
        """
        Run date

        :return: datetime.date, Run date
        """
        if cls._run_date is None:
            raise ConfigException("Run Date not Set !")
        return cls._run_date

    @classmethod
    def set_feature_store_config(cls, feature_store_schema: str, feature_metadata_schema: str) -> None:
        """
        Inicialize feature store config

        :param feature_store_schema: str, schema name
        :param feature_metadata_schema: str, metadata schema name
        :return: None
        """
        cls._feature_store_config = FeatureStoreConfig(
            feature_store_schema=feature_store_schema, feature_metadata_schema=feature_metadata_schema
        )

    @classmethod
    def get_feature_store_config(cls) -> FeatureStoreConfig:
        """
        Feature Store Config

        :return: FeatureStoreConfig
        """
        if cls._feature_store_config is None:
            raise ConfigException("Feature Store Config not Set !")

        return cls._feature_store_config

    @classmethod
    def get_config(cls) -> typing.Dict:
        """
        Get config dictionary

        :return: dictionary of key-value pairs
        """
        return cls._config.copy()

    @classmethod
    def set_custom_config(cls, **kwargs) -> None:
        """
        Set custom key-value pairs for custom config

        :param kwargs: key-value pairs to setup
        :return: None
        """
        cls._config.update(kwargs)

    @classmethod
    def get_dependency_config(cls) -> typing.Dict:
        """
        Get rialto job dependency config

        :return: dictionary with dependency config
        """
        return cls._dependencies

    @classmethod
    def set_dependency_config(cls, dependencies: typing.Dict) -> None:
        """
        Get rialto job dependency config

        :param dependencies: dictionary with the config
        :return: None
        """
        cls._dependencies = dependencies



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["JobBase"]

import abc
import datetime
import typing
from contextlib import contextmanager

import pyspark.sql.functions as F
from loguru import logger
from pyspark.sql import DataFrame, SparkSession

from rialto.common import TableReader
from rialto.jobs.configuration.config_holder import ConfigHolder
from rialto.jobs.decorators.resolver import Resolver
from rialto.loader import DatabricksLoader, PysparkFeatureLoader
from rialto.metadata import MetadataManager
from rialto.runner import Transformation


class JobBase(Transformation):
    """A Base Class for Rialto Jobs. Serves as a foundation into which the @job decorators are inserted."""

    @abc.abstractmethod
    def get_custom_callable(self) -> typing.Callable:
        """Getter - Custom callable (i.e. job transformation function)"""
        pass

    @abc.abstractmethod
    def get_job_version(self) -> str:
        """Job version getter"""
        pass

    @abc.abstractmethod
    def get_job_name(self) -> str:
        """Job name getter"""
        pass

    @contextmanager
    def _setup_resolver(self, run_date: datetime.date) -> None:
        Resolver.register_callable(lambda: run_date, "run_date")

        Resolver.register_callable(ConfigHolder.get_config, "config")
        Resolver.register_callable(ConfigHolder.get_dependency_config, "dependencies")

        Resolver.register_callable(self._get_spark, "spark")
        Resolver.register_callable(self._get_table_reader, "table_reader")
        Resolver.register_callable(self._get_feature_loader, "feature_loader")

        try:
            yield
        finally:
            Resolver.cache_clear()

    def _setup(
        self, spark: SparkSession, run_date: datetime.date, table_reader: TableReader, dependencies: typing.Dict = None
    ) -> None:
        self._spark = spark
        self._table_rader = table_reader

        ConfigHolder.set_dependency_config(dependencies)
        ConfigHolder.set_run_date(run_date)

    def _get_spark(self) -> SparkSession:
        return self._spark

    def _get_table_reader(self) -> TableReader:
        return self._table_rader

    def _get_feature_loader(self) -> PysparkFeatureLoader:
        config = ConfigHolder.get_feature_store_config()

        databricks_loader = DatabricksLoader(self._spark, config.feature_store_schema)
        feature_loader = PysparkFeatureLoader(self._spark, databricks_loader, config.feature_metadata_schema)

        return feature_loader

    def _get_timestamp_holder_result(self) -> DataFrame:
        spark = self._get_spark()
        return spark.createDataFrame(
            [(self.get_job_name(), datetime.datetime.now())], schema="JOB_NAME string, CREATION_TIME timestamp"
        )

    def _add_job_version(self, df: DataFrame) -> DataFrame:
        version = self.get_job_version()
        return df.withColumn("VERSION", F.lit(version))

    def _run_main_callable(self, run_date: datetime.date) -> DataFrame:
        with self._setup_resolver(run_date):
            custom_callable = self.get_custom_callable()
            raw_result = Resolver.register_resolve(custom_callable)

        if raw_result is None:
            raw_result = self._get_timestamp_holder_result()

        result_with_version = self._add_job_version(raw_result)
        return result_with_version

    def run(
        self,
        reader: TableReader,
        run_date: datetime.date,
        spark: SparkSession = None,
        metadata_manager: MetadataManager = None,
        dependencies: typing.Dict = None,
    ) -> DataFrame:
        """
        Rialto transformation run

        :param reader: data store api object
        :param info_date: date
        :param spark: spark session
        :param metadata_manager: metadata api object
        :param dependencies: rialto job dependencies
        :return: dataframe
        """
        try:
            self._setup(spark, run_date, reader, dependencies)
            return self._run_main_callable(run_date)
        except Exception as e:
            logger.exception(e)
            raise e



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["load_yaml", "get_date_col_property", "get_delta_partition"]

import os
from typing import Any

import pyspark.sql.functions as F
import yaml
from pyspark.sql import DataFrame
from pyspark.sql.types import FloatType


def load_yaml(path: str) -> Any:
    """
    YAML loader

    :param path: file path
    :return: Parsed yaml
    """
    if not os.path.isfile(path):
        raise FileNotFoundError(f"Can't find {path}.")

    with open(path, "r") as stream:
        return yaml.safe_load(stream)


def get_date_col_property(spark, table: str, property: str) -> str:
    """
    Retrieve a data column name from a given table property

    :param spark: spark session
    :param table: path to table
    :param property: name of the property
    :return: data column name
    """
    props = spark.sql(f"show tblproperties {table}")
    date_col = props.filter(F.col("key") == property).select("value").collect()
    if len(date_col):
        return date_col[0].value
    else:
        raise RuntimeError(f"Table {table} has no property {property}.")


def get_delta_partition(spark, table: str) -> str:
    """
    Select first partition column of the delta table

    :param table: full table name
    :return: partition column name
    """
    columns = spark.catalog.listColumns(table)
    partition_columns = list(filter(lambda c: c.isPartition, columns))
    if len(partition_columns):
        return partition_columns[0].name
    else:
        raise RuntimeError(f"Delta table has no partitions: {table}.")


def cast_decimals_to_floats(df: DataFrame) -> DataFrame:
    """
    Find all decimal types in the table and cast them to floats. Fixes errors in .toPandas() conversions.

    :param df: pyspark DataFrame
    :return: pyspark DataFrame with fixed types
    """
    decimal_cols = [col_name for col_name, data_type in df.dtypes if "decimal" in data_type]
    for c in decimal_cols:
        df = df.withColumn(c, F.col(c).cast(FloatType()))

    return df



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["ResolverException", "Resolver"]

import inspect
import typing
from functools import cache


class ResolverException(Exception):
    """Resolver Errors Class - In Most Cases your dependency tree is not complete."""

    pass


class Resolver:
    """
    Resolver handles dependency management between datasets and jobs.

    We register different callables, which can depend on other callables.
    Calling resolve() we attempts to resolve these dependencies.
    """

    _storage = {}

    @classmethod
    def _get_args_for_call(cls, function: typing.Callable) -> typing.Dict[str, typing.Any]:
        result_dict = {}
        signature = inspect.signature(function)

        for param in signature.parameters.values():
            result_dict[param.name] = cls.resolve(param.name)

        return result_dict

    @classmethod
    def register_callable(cls, callable: typing.Callable, name: str = None) -> str:
        """
        Register callable with a given name for later resolution.

        In case name isn't present, function's __name__ attribute will be used.

        :param callable: callable to register (getter)
        :param name: str, custom name, f.__name__ will be used otherwise
        :return: str, name under which the callable has been registered
        """
        if name is None:
            name = getattr(callable, "__name__", repr(callable))

        cls._storage[name] = callable
        return name

    @classmethod
    @cache
    def resolve(cls, name: str) -> typing.Any:
        """
        Search for a callable registered prior and attempt to call it with correct arguents.

        Arguments are resolved recursively according to requirements; For example, if we have
        a(b, c), b(d), and c(), d() registered, then we recursively call resolve() methods until we resolve
        c, d -> b -> a

        :param name: name of the callable to resolve
        :return: result of the callable
        """
        if name not in cls._storage.keys():
            raise ResolverException(f"{name} declaration not found!")

        getter = cls._storage[name]
        args = cls._get_args_for_call(getter)

        return getter(**args)

    @classmethod
    def register_resolve(cls, callable: typing.Callable) -> typing.Any:
        """
        Register and Resolve a callable.

        Combination of the register() and resolve() methods for a simplified execution.

        :param callable: callable to register and immediately resolve
        :return: result of the callable
        """
        name = cls.register_callable(callable)
        return cls.resolve(name)

    @classmethod
    def cache_clear(cls) -> None:
        """
        Clear resolver cache.

        The resolve mehtod caches its results to avoid duplication of resolutions.
        However, in case we re-register some callables, we need to clear cache
        in order to ensure re-execution of all resolutions.

        :return: None
        """
        cls.resolve.cache_clear()
